import net.sf.javabdd.*;
import java.util.*;
import java.lang.reflect.Array;
import org.python.core.PySlice;

public class SymBDD {
  private String name;
  public String name() { return(name); }

  private SymFactory sf;
  public SymFactory sf() { return(sf); }

  private int width;
  public int width() { return(width); }

  protected SymBDD(String name, int width, SymFactory sf) {
    init_SymBDD(sf);
    if(name == null) {
      if(width < 0)
        throw new IllegalArgumentException("SymBDD(constructor) --  bad width:  " + width);
      if(width == 1)
	name = (new Formatter()).format("$%03d", sf.varNum()).toString();
      else
	name = (new Formatter()).format("$%03d:%03d", sf.varNum(), sf.varNum()+width-1).toString();
    } else {
      // Legal names are sequences of letters, digits, and underscores,
      // that start with a letter or digit.  If this SymBDD has width()=1,
      // then it may be of the form Identifier[Integer], where Identifier
      // is a sequence of letters, digits, and underscores, and Integer
      // is a sequence of digits, and Identifier is not otherwise defined.
      for(int i = 0; i < name.length(); i++) {
        char c = name.charAt(i);
        if(Character.isLetterOrDigit(c) || (c == '_')) continue;
	if((c == '[') &&  (width == 1)) {
	  int j;
	  for(j = i+1; (j < name.length()) && Character.isDigit(name.charAt(j)); j++);
	  if(j == name.length()-1 && (name.charAt(j) == ']')) break;
	  throw new IllegalArgumentException("missing ']' in SymBDD name -- " + name);
	}
	throw new IllegalArgumentException("illegal character in SymBDD name -- " + c);
      }
    }
    this.name = name;
    this.width = width;
    this.sf.putSym(this);
  }
  protected SymBDD(int nbits, SymFactory sf) {
    init_SymBDD(sf);
    this.name = null;
    this.width = nbits;
  }

  private void init_SymBDD(SymFactory sf) {
    if(sf != null) this.sf = sf;
    else this.sf = SymFactory.defaultSF();
  }

  public SymBDD find(String name) {
    return(sf().do_find(name));
  }

  public BDD getBDD() {
    throw new IllegalArgumentException("SymBDD.getBool(): I'm not a SymBool");
  }

  public BDD getBDD(int i) {
    throw new IllegalArgumentException("SymBDD.getBool(): I'm not a SymInt or SymUInt");
  }

  // The following methods can be overridded by subclasses
  // (to do something useful)
  public SymBDD nope(String op, Object that) {
    throw new UnsupportedOperationException(
      "this." + op + "(that) not supported for:\n"
      + "  " + this.getClass() + " this;\n"
      + "  " + that.getClass() + " that;");
  }
  public SymBDD nope(String op) {
    throw new UnsupportedOperationException(
      "this." + op + "() not supported for " + this.getClass());
  }
  public SymBDD huh(String op, SymBDD that) {
    throw new UnsupportedOperationException(
      "I don't know how to sanely implement " +
      "this." + op + "(that) for:\n"
      + "  " + this.getClass() + " this;\n"
      + "  " + that.getClass() + " that;");
  }

  public SymBDD abs() { return(this); }
  public SymBDD neg() { return(nope("neg")); }
  public SymBDD not() { return(nope("not")); }

  // functions for overloading python operators
  // We don't implement __cmp__, for each valuation of a symbolic expression
  // __cmp__ would need to return -1, 0, or +1, which would double our number
  // of variables and generally mess things up.
  // Instead, we provide the "deep" comparison operations: __eq__, etc.
  public SymBDD __abs__()  	           { return(abs()); }
  public SymBDD __invert__()  	           { return(not()); }

  // The code for binary operators is *really* boring.
  // It is produced by SymBDD.Dispatch.main()
  public static class Dispatch {
    public static void main(String[] args) {
      String[] argtype = new String[] { 
        "boolean", "int[]", "int", "Object[]",
	"SymBool", "SymBool[]", "SymUInt"
      };
      Formatter out = new Formatter(System.out);
      for(SymOp s : SymOp.values()) {
        for(String p : s.pyNames()) {
	  for(String a : argtype) {
	    out.format("  public SymBDD __%s__(%s that)", p, a);
	    out.format("{ return(op(that, SymOp.%s)); }\n", s.toString());
	  }
	  out.format("\n");
	}
      }
      out.flush();
    }
  }

  // begin (code produced by SymBDD.Dispatch.main)
  public SymBDD __add__(boolean that){ return(op(that, SymOp.ADD)); }
  public SymBDD __add__(int[] that){ return(op(that, SymOp.ADD)); }
  public SymBDD __add__(int that){ return(op(that, SymOp.ADD)); }
  public SymBDD __add__(Object[] that){ return(op(that, SymOp.ADD)); }
  public SymBDD __add__(SymBool that){ return(op(that, SymOp.ADD)); }
  public SymBDD __add__(SymBool[] that){ return(op(that, SymOp.ADD)); }
  public SymBDD __add__(SymUInt that){ return(op(that, SymOp.ADD)); }

  public SymBDD __radd__(boolean that){ return(op(that, SymOp.ADD)); }
  public SymBDD __radd__(int[] that){ return(op(that, SymOp.ADD)); }
  public SymBDD __radd__(int that){ return(op(that, SymOp.ADD)); }
  public SymBDD __radd__(Object[] that){ return(op(that, SymOp.ADD)); }
  public SymBDD __radd__(SymBool that){ return(op(that, SymOp.ADD)); }
  public SymBDD __radd__(SymBool[] that){ return(op(that, SymOp.ADD)); }
  public SymBDD __radd__(SymUInt that){ return(op(that, SymOp.ADD)); }

  public SymBDD __and__(boolean that){ return(op(that, SymOp.AND)); }
  public SymBDD __and__(int[] that){ return(op(that, SymOp.AND)); }
  public SymBDD __and__(int that){ return(op(that, SymOp.AND)); }
  public SymBDD __and__(Object[] that){ return(op(that, SymOp.AND)); }
  public SymBDD __and__(SymBool that){ return(op(that, SymOp.AND)); }
  public SymBDD __and__(SymBool[] that){ return(op(that, SymOp.AND)); }
  public SymBDD __and__(SymUInt that){ return(op(that, SymOp.AND)); }

  public SymBDD __rand__(boolean that){ return(op(that, SymOp.AND)); }
  public SymBDD __rand__(int[] that){ return(op(that, SymOp.AND)); }
  public SymBDD __rand__(int that){ return(op(that, SymOp.AND)); }
  public SymBDD __rand__(Object[] that){ return(op(that, SymOp.AND)); }
  public SymBDD __rand__(SymBool that){ return(op(that, SymOp.AND)); }
  public SymBDD __rand__(SymBool[] that){ return(op(that, SymOp.AND)); }
  public SymBDD __rand__(SymUInt that){ return(op(that, SymOp.AND)); }

  public SymBDD __div__(boolean that){ return(op(that, SymOp.DIV)); }
  public SymBDD __div__(int[] that){ return(op(that, SymOp.DIV)); }
  public SymBDD __div__(int that){ return(op(that, SymOp.DIV)); }
  public SymBDD __div__(Object[] that){ return(op(that, SymOp.DIV)); }
  public SymBDD __div__(SymBool that){ return(op(that, SymOp.DIV)); }
  public SymBDD __div__(SymBool[] that){ return(op(that, SymOp.DIV)); }
  public SymBDD __div__(SymUInt that){ return(op(that, SymOp.DIV)); }

  public SymBDD __eq__(boolean that){ return(op(that, SymOp.EQ)); }
  public SymBDD __eq__(int[] that){ return(op(that, SymOp.EQ)); }
  public SymBDD __eq__(int that){ return(op(that, SymOp.EQ)); }
  public SymBDD __eq__(Object[] that){ return(op(that, SymOp.EQ)); }
  public SymBDD __eq__(SymBool that){ return(op(that, SymOp.EQ)); }
  public SymBDD __eq__(SymBool[] that){ return(op(that, SymOp.EQ)); }
  public SymBDD __eq__(SymUInt that){ return(op(that, SymOp.EQ)); }

  public SymBDD __req__(boolean that){ return(op(that, SymOp.EQ)); }
  public SymBDD __req__(int[] that){ return(op(that, SymOp.EQ)); }
  public SymBDD __req__(int that){ return(op(that, SymOp.EQ)); }
  public SymBDD __req__(Object[] that){ return(op(that, SymOp.EQ)); }
  public SymBDD __req__(SymBool that){ return(op(that, SymOp.EQ)); }
  public SymBDD __req__(SymBool[] that){ return(op(that, SymOp.EQ)); }
  public SymBDD __req__(SymUInt that){ return(op(that, SymOp.EQ)); }

  public SymBDD __ge__(boolean that){ return(op(that, SymOp.GE)); }
  public SymBDD __ge__(int[] that){ return(op(that, SymOp.GE)); }
  public SymBDD __ge__(int that){ return(op(that, SymOp.GE)); }
  public SymBDD __ge__(Object[] that){ return(op(that, SymOp.GE)); }
  public SymBDD __ge__(SymBool that){ return(op(that, SymOp.GE)); }
  public SymBDD __ge__(SymBool[] that){ return(op(that, SymOp.GE)); }
  public SymBDD __ge__(SymUInt that){ return(op(that, SymOp.GE)); }

  public SymBDD __rle__(boolean that){ return(op(that, SymOp.GE)); }
  public SymBDD __rle__(int[] that){ return(op(that, SymOp.GE)); }
  public SymBDD __rle__(int that){ return(op(that, SymOp.GE)); }
  public SymBDD __rle__(Object[] that){ return(op(that, SymOp.GE)); }
  public SymBDD __rle__(SymBool that){ return(op(that, SymOp.GE)); }
  public SymBDD __rle__(SymBool[] that){ return(op(that, SymOp.GE)); }
  public SymBDD __rle__(SymUInt that){ return(op(that, SymOp.GE)); }

  public SymBDD __gt__(boolean that){ return(op(that, SymOp.GT)); }
  public SymBDD __gt__(int[] that){ return(op(that, SymOp.GT)); }
  public SymBDD __gt__(int that){ return(op(that, SymOp.GT)); }
  public SymBDD __gt__(Object[] that){ return(op(that, SymOp.GT)); }
  public SymBDD __gt__(SymBool that){ return(op(that, SymOp.GT)); }
  public SymBDD __gt__(SymBool[] that){ return(op(that, SymOp.GT)); }
  public SymBDD __gt__(SymUInt that){ return(op(that, SymOp.GT)); }

  public SymBDD __rlt__(boolean that){ return(op(that, SymOp.GT)); }
  public SymBDD __rlt__(int[] that){ return(op(that, SymOp.GT)); }
  public SymBDD __rlt__(int that){ return(op(that, SymOp.GT)); }
  public SymBDD __rlt__(Object[] that){ return(op(that, SymOp.GT)); }
  public SymBDD __rlt__(SymBool that){ return(op(that, SymOp.GT)); }
  public SymBDD __rlt__(SymBool[] that){ return(op(that, SymOp.GT)); }
  public SymBDD __rlt__(SymUInt that){ return(op(that, SymOp.GT)); }

  public SymBDD __le__(boolean that){ return(op(that, SymOp.LE)); }
  public SymBDD __le__(int[] that){ return(op(that, SymOp.LE)); }
  public SymBDD __le__(int that){ return(op(that, SymOp.LE)); }
  public SymBDD __le__(Object[] that){ return(op(that, SymOp.LE)); }
  public SymBDD __le__(SymBool that){ return(op(that, SymOp.LE)); }
  public SymBDD __le__(SymBool[] that){ return(op(that, SymOp.LE)); }
  public SymBDD __le__(SymUInt that){ return(op(that, SymOp.LE)); }

  public SymBDD __rge__(boolean that){ return(op(that, SymOp.LE)); }
  public SymBDD __rge__(int[] that){ return(op(that, SymOp.LE)); }
  public SymBDD __rge__(int that){ return(op(that, SymOp.LE)); }
  public SymBDD __rge__(Object[] that){ return(op(that, SymOp.LE)); }
  public SymBDD __rge__(SymBool that){ return(op(that, SymOp.LE)); }
  public SymBDD __rge__(SymBool[] that){ return(op(that, SymOp.LE)); }
  public SymBDD __rge__(SymUInt that){ return(op(that, SymOp.LE)); }

  public SymBDD __lt__(boolean that){ return(op(that, SymOp.LT)); }
  public SymBDD __lt__(int[] that){ return(op(that, SymOp.LT)); }
  public SymBDD __lt__(int that){ return(op(that, SymOp.LT)); }
  public SymBDD __lt__(Object[] that){ return(op(that, SymOp.LT)); }
  public SymBDD __lt__(SymBool that){ return(op(that, SymOp.LT)); }
  public SymBDD __lt__(SymBool[] that){ return(op(that, SymOp.LT)); }
  public SymBDD __lt__(SymUInt that){ return(op(that, SymOp.LT)); }

  public SymBDD __rgt__(boolean that){ return(op(that, SymOp.LT)); }
  public SymBDD __rgt__(int[] that){ return(op(that, SymOp.LT)); }
  public SymBDD __rgt__(int that){ return(op(that, SymOp.LT)); }
  public SymBDD __rgt__(Object[] that){ return(op(that, SymOp.LT)); }
  public SymBDD __rgt__(SymBool that){ return(op(that, SymOp.LT)); }
  public SymBDD __rgt__(SymBool[] that){ return(op(that, SymOp.LT)); }
  public SymBDD __rgt__(SymUInt that){ return(op(that, SymOp.LT)); }

  public SymBDD __mod__(boolean that){ return(op(that, SymOp.MOD)); }
  public SymBDD __mod__(int[] that){ return(op(that, SymOp.MOD)); }
  public SymBDD __mod__(int that){ return(op(that, SymOp.MOD)); }
  public SymBDD __mod__(Object[] that){ return(op(that, SymOp.MOD)); }
  public SymBDD __mod__(SymBool that){ return(op(that, SymOp.MOD)); }
  public SymBDD __mod__(SymBool[] that){ return(op(that, SymOp.MOD)); }
  public SymBDD __mod__(SymUInt that){ return(op(that, SymOp.MOD)); }

  public SymBDD __mul__(boolean that){ return(op(that, SymOp.MUL)); }
  public SymBDD __mul__(int[] that){ return(op(that, SymOp.MUL)); }
  public SymBDD __mul__(int that){ return(op(that, SymOp.MUL)); }
  public SymBDD __mul__(Object[] that){ return(op(that, SymOp.MUL)); }
  public SymBDD __mul__(SymBool that){ return(op(that, SymOp.MUL)); }
  public SymBDD __mul__(SymBool[] that){ return(op(that, SymOp.MUL)); }
  public SymBDD __mul__(SymUInt that){ return(op(that, SymOp.MUL)); }

  public SymBDD __rmul__(boolean that){ return(op(that, SymOp.MUL)); }
  public SymBDD __rmul__(int[] that){ return(op(that, SymOp.MUL)); }
  public SymBDD __rmul__(int that){ return(op(that, SymOp.MUL)); }
  public SymBDD __rmul__(Object[] that){ return(op(that, SymOp.MUL)); }
  public SymBDD __rmul__(SymBool that){ return(op(that, SymOp.MUL)); }
  public SymBDD __rmul__(SymBool[] that){ return(op(that, SymOp.MUL)); }
  public SymBDD __rmul__(SymUInt that){ return(op(that, SymOp.MUL)); }

  public SymBDD __ne__(boolean that){ return(op(that, SymOp.NE)); }
  public SymBDD __ne__(int[] that){ return(op(that, SymOp.NE)); }
  public SymBDD __ne__(int that){ return(op(that, SymOp.NE)); }
  public SymBDD __ne__(Object[] that){ return(op(that, SymOp.NE)); }
  public SymBDD __ne__(SymBool that){ return(op(that, SymOp.NE)); }
  public SymBDD __ne__(SymBool[] that){ return(op(that, SymOp.NE)); }
  public SymBDD __ne__(SymUInt that){ return(op(that, SymOp.NE)); }

  public SymBDD __rne__(boolean that){ return(op(that, SymOp.NE)); }
  public SymBDD __rne__(int[] that){ return(op(that, SymOp.NE)); }
  public SymBDD __rne__(int that){ return(op(that, SymOp.NE)); }
  public SymBDD __rne__(Object[] that){ return(op(that, SymOp.NE)); }
  public SymBDD __rne__(SymBool that){ return(op(that, SymOp.NE)); }
  public SymBDD __rne__(SymBool[] that){ return(op(that, SymOp.NE)); }
  public SymBDD __rne__(SymUInt that){ return(op(that, SymOp.NE)); }

  public SymBDD __or__(boolean that){ return(op(that, SymOp.OR)); }
  public SymBDD __or__(int[] that){ return(op(that, SymOp.OR)); }
  public SymBDD __or__(int that){ return(op(that, SymOp.OR)); }
  public SymBDD __or__(Object[] that){ return(op(that, SymOp.OR)); }
  public SymBDD __or__(SymBool that){ return(op(that, SymOp.OR)); }
  public SymBDD __or__(SymBool[] that){ return(op(that, SymOp.OR)); }
  public SymBDD __or__(SymUInt that){ return(op(that, SymOp.OR)); }

  public SymBDD __ror__(boolean that){ return(op(that, SymOp.OR)); }
  public SymBDD __ror__(int[] that){ return(op(that, SymOp.OR)); }
  public SymBDD __ror__(int that){ return(op(that, SymOp.OR)); }
  public SymBDD __ror__(Object[] that){ return(op(that, SymOp.OR)); }
  public SymBDD __ror__(SymBool that){ return(op(that, SymOp.OR)); }
  public SymBDD __ror__(SymBool[] that){ return(op(that, SymOp.OR)); }
  public SymBDD __ror__(SymUInt that){ return(op(that, SymOp.OR)); }

  public SymBDD __lshift__(boolean that){ return(op(that, SymOp.SHL)); }
  public SymBDD __lshift__(int[] that){ return(op(that, SymOp.SHL)); }
  public SymBDD __lshift__(int that){ return(op(that, SymOp.SHL)); }
  public SymBDD __lshift__(Object[] that){ return(op(that, SymOp.SHL)); }
  public SymBDD __lshift__(SymBool that){ return(op(that, SymOp.SHL)); }
  public SymBDD __lshift__(SymBool[] that){ return(op(that, SymOp.SHL)); }
  public SymBDD __lshift__(SymUInt that){ return(op(that, SymOp.SHL)); }

  public SymBDD __rshift__(boolean that){ return(op(that, SymOp.SHR)); }
  public SymBDD __rshift__(int[] that){ return(op(that, SymOp.SHR)); }
  public SymBDD __rshift__(int that){ return(op(that, SymOp.SHR)); }
  public SymBDD __rshift__(Object[] that){ return(op(that, SymOp.SHR)); }
  public SymBDD __rshift__(SymBool that){ return(op(that, SymOp.SHR)); }
  public SymBDD __rshift__(SymBool[] that){ return(op(that, SymOp.SHR)); }
  public SymBDD __rshift__(SymUInt that){ return(op(that, SymOp.SHR)); }

  public SymBDD __sub__(boolean that){ return(op(that, SymOp.SUB)); }
  public SymBDD __sub__(int[] that){ return(op(that, SymOp.SUB)); }
  public SymBDD __sub__(int that){ return(op(that, SymOp.SUB)); }
  public SymBDD __sub__(Object[] that){ return(op(that, SymOp.SUB)); }
  public SymBDD __sub__(SymBool that){ return(op(that, SymOp.SUB)); }
  public SymBDD __sub__(SymBool[] that){ return(op(that, SymOp.SUB)); }
  public SymBDD __sub__(SymUInt that){ return(op(that, SymOp.SUB)); }

  public SymBDD __xor__(boolean that){ return(op(that, SymOp.XOR)); }
  public SymBDD __xor__(int[] that){ return(op(that, SymOp.XOR)); }
  public SymBDD __xor__(int that){ return(op(that, SymOp.XOR)); }
  public SymBDD __xor__(Object[] that){ return(op(that, SymOp.XOR)); }
  public SymBDD __xor__(SymBool that){ return(op(that, SymOp.XOR)); }
  public SymBDD __xor__(SymBool[] that){ return(op(that, SymOp.XOR)); }
  public SymBDD __xor__(SymUInt that){ return(op(that, SymOp.XOR)); }

  public SymBDD __rxor__(boolean that){ return(op(that, SymOp.XOR)); }
  public SymBDD __rxor__(int[] that){ return(op(that, SymOp.XOR)); }
  public SymBDD __rxor__(int that){ return(op(that, SymOp.XOR)); }
  public SymBDD __rxor__(Object[] that){ return(op(that, SymOp.XOR)); }
  public SymBDD __rxor__(SymBool that){ return(op(that, SymOp.XOR)); }
  public SymBDD __rxor__(SymBool[] that){ return(op(that, SymOp.XOR)); }
  public SymBDD __rxor__(SymUInt that){ return(op(that, SymOp.XOR)); }

  public SymBDD __rdiv__(boolean that){ return(op(that, SymOp.RDIV)); }
  public SymBDD __rdiv__(int[] that){ return(op(that, SymOp.RDIV)); }
  public SymBDD __rdiv__(int that){ return(op(that, SymOp.RDIV)); }
  public SymBDD __rdiv__(Object[] that){ return(op(that, SymOp.RDIV)); }
  public SymBDD __rdiv__(SymBool that){ return(op(that, SymOp.RDIV)); }
  public SymBDD __rdiv__(SymBool[] that){ return(op(that, SymOp.RDIV)); }
  public SymBDD __rdiv__(SymUInt that){ return(op(that, SymOp.RDIV)); }

  public SymBDD __rimp__(boolean that){ return(op(that, SymOp.RIMP)); }
  public SymBDD __rimp__(int[] that){ return(op(that, SymOp.RIMP)); }
  public SymBDD __rimp__(int that){ return(op(that, SymOp.RIMP)); }
  public SymBDD __rimp__(Object[] that){ return(op(that, SymOp.RIMP)); }
  public SymBDD __rimp__(SymBool that){ return(op(that, SymOp.RIMP)); }
  public SymBDD __rimp__(SymBool[] that){ return(op(that, SymOp.RIMP)); }
  public SymBDD __rimp__(SymUInt that){ return(op(that, SymOp.RIMP)); }

  public SymBDD __rmod__(boolean that){ return(op(that, SymOp.RMOD)); }
  public SymBDD __rmod__(int[] that){ return(op(that, SymOp.RMOD)); }
  public SymBDD __rmod__(int that){ return(op(that, SymOp.RMOD)); }
  public SymBDD __rmod__(Object[] that){ return(op(that, SymOp.RMOD)); }
  public SymBDD __rmod__(SymBool that){ return(op(that, SymOp.RMOD)); }
  public SymBDD __rmod__(SymBool[] that){ return(op(that, SymOp.RMOD)); }
  public SymBDD __rmod__(SymUInt that){ return(op(that, SymOp.RMOD)); }

  public SymBDD __rlshift__(boolean that){ return(op(that, SymOp.RSHL)); }
  public SymBDD __rlshift__(int[] that){ return(op(that, SymOp.RSHL)); }
  public SymBDD __rlshift__(int that){ return(op(that, SymOp.RSHL)); }
  public SymBDD __rlshift__(Object[] that){ return(op(that, SymOp.RSHL)); }
  public SymBDD __rlshift__(SymBool that){ return(op(that, SymOp.RSHL)); }
  public SymBDD __rlshift__(SymBool[] that){ return(op(that, SymOp.RSHL)); }
  public SymBDD __rlshift__(SymUInt that){ return(op(that, SymOp.RSHL)); }

  public SymBDD __rrshift__(boolean that){ return(op(that, SymOp.RSHR)); }
  public SymBDD __rrshift__(int[] that){ return(op(that, SymOp.RSHR)); }
  public SymBDD __rrshift__(int that){ return(op(that, SymOp.RSHR)); }
  public SymBDD __rrshift__(Object[] that){ return(op(that, SymOp.RSHR)); }
  public SymBDD __rrshift__(SymBool that){ return(op(that, SymOp.RSHR)); }
  public SymBDD __rrshift__(SymBool[] that){ return(op(that, SymOp.RSHR)); }
  public SymBDD __rrshift__(SymUInt that){ return(op(that, SymOp.RSHR)); }

  public SymBDD __rsub__(boolean that){ return(op(that, SymOp.RSUB)); }
  public SymBDD __rsub__(int[] that){ return(op(that, SymOp.RSUB)); }
  public SymBDD __rsub__(int that){ return(op(that, SymOp.RSUB)); }
  public SymBDD __rsub__(Object[] that){ return(op(that, SymOp.RSUB)); }
  public SymBDD __rsub__(SymBool that){ return(op(that, SymOp.RSUB)); }
  public SymBDD __rsub__(SymBool[] that){ return(op(that, SymOp.RSUB)); }
  public SymBDD __rsub__(SymUInt that){ return(op(that, SymOp.RSUB)); }
  // end (code produced by SymBDD.Dispatch.main)


  public SymBDD op(SymBool that, SymOp op) {
    throw new UnsupportedOperationException();
  }
  public SymBDD op(SymUInt that, SymOp op) {
    throw new UnsupportedOperationException();
  }
  public SymBDD op(int that, SymOp op) {
    throw new UnsupportedOperationException();
  }
  public SymBDD op(boolean that, SymOp op) { return(op(mkSB(that), op)); }
  public SymBDD op(int[] that, SymOp op) { return(op(mkSUI(that), op)); }
  public SymBDD op(SymBool[] that, SymOp op) { return(op(mkSUI(that), op)); }
  public SymBDD op(Object[] that, SymOp op)  { return(op(mkSUI(that), op)); }

  public static void main(String[] args) {
    Dispatch.main(args);
  }

  // Now we'll handle the various ways to invoke exist() and forAll()
  public SymBDD exist(SymBool[] vars) {  // should be overridden subclasses
    throw new UnsupportedOperationException();
  }
  public SymBDD exist(SymBool v)  { return(exist(getVars(v, "exist"))); }
  public SymBDD exist(SymUInt v)  { return(exist(getVars(v, "exist"))); }
  public SymBDD exist(Object[] v) { return(exist(getVars(v, "exist"))); }

  public SymBDD forAll(SymBool[] vars) {  // should be overridden subclasses
    throw new UnsupportedOperationException();
  }
  public SymBDD forAll(SymBool v)  { return(forAll(getVars(v, "forAll"))); }
  public SymBDD forAll(SymUInt v)  { return(forAll(getVars(v, "forAll"))); }
  public SymBDD forAll(Object[] v) { return(forAll(getVars(v, "forAll"))); }


  public SymBDD relProd(SymBool that, SymBool[] vars) {
    // this method should be overridden by any subclass that supports relProd()
    throw new UnsupportedOperationException();
  }
  public SymBDD relProd(int that, SymBool[] vars) {
    return(relProd(sf().bool((that & 1) == 1), vars));
  }
  public SymBDD relProd(SymBool that, SymUInt vars) {
    return(relProd(that, vars.toArray()));
  }
  public SymBDD relProd(int that, SymUInt vars) {
    return(relProd(sf().bool((that & 1) == 1), vars.toArray()));
  }


  // The compose operator also has lots of ways that the Python might
  //   invoke it.  Here's the dispatch code
  // compose(oldVars, newExprs) should be overridden by any subclass
  // of SymBDD (e.g. SymBool and SymUInt) for which composition makes
  // sense.
  public SymBDD compose(SymBool[] oldVars, SymBool[] newExprs) {
    throw new UnsupportedOperationException();
  }
  public SymBDD comp(Object oV, Object nE) {
    List<SymBool> vList = new LinkedList<SymBool>();
    List<SymBool> eList = new LinkedList<SymBool>();
    comp(oV, nE, vList, eList);
    SymBool[] oldVars = vList.toArray(new SymBool[vList.size()]);
    SymBool[] newExprs = eList.toArray(new SymBool[eList.size()]);
    return(compose(oldVars, newExprs));
  }
  public void comp(Object oV, Object nE, List<SymBool> vList, List<SymBool> eList) {
    if(oV instanceof SymBool) {
      SymBool s = (SymBool)(oV);
      if(!(s.isVar()))
        throw new UnsupportedOperationException("compose: first argument must be a variable");
      eList.add(mkSB(nE));
      vList.add(s);
    } else if(oV instanceof SymBool[]) {
      // treated the same as a SymUInt of the same width
      comp(new SymUInt((SymBool[])(oV), sf()), nE, vList, eList);
    } else if(oV instanceof SymUInt) {
      SymUInt ui = (SymUInt)(oV);
      SymBool[] aV = ui.toArray();
      SymBool[] aE = ui.mkSBA(nE);
      for(int i = 0; i < aV.length; i++) {
        vList.add(aV[i]);
	eList.add(aE[i]);
      }
    } else if(oV.getClass().isArray()) {
      // first see if oV is an array of SymBool variables.
      // if so, convert oV to a SymUInt and recurse.
      // Otherwise, nE must also be an array of the same length,
      //   and we recurse on corresponding elements.
      boolean isSBA = true;
      for(int i = 0; isSBA && (i < Array.getLength(oV)); i++)
        isSBA &= Array.get(oV, i) instanceof SymBool;
      if(isSBA) {
        SymBool[] a = new SymBool[Array.getLength(oV)];
	for(int i = 0; i < a.length; i++)
	  a[i] = (SymBool)(Array.get(oV, i));
        comp(new SymUInt(a, sf()), nE, vList, eList);
      } else if(!nE.getClass().isArray()) {
        throw new IllegalArgumentException(
	  "compose: can't match variables with expressions");
      } else if(Array.getLength(oV) != Array.getLength(nE)) {
        throw new IllegalArgumentException(
	  "compose: mismatched arrays for variables and expressions");
      } else {
        for(int i = 0; i < Array.getLength(oV); i++)
	  comp(Array.get(oV, i), Array.get(nE, i), vList, eList);
      }
    }
  }
  public SymBDD compose(SymBool oV, SymBool nE)     { return(comp(oV, nE)); }
  public SymBDD compose(SymBool oV, SymBool[] nE)   { return(comp(oV, nE)); }
  public SymBDD compose(SymBool oV, SymUInt nE)     { return(comp(oV, nE)); }
  public SymBDD compose(SymBool oV, int[] nE)       { return(comp(oV, nE)); }
  public SymBDD compose(SymBool oV, Integer nE)     { return(comp(oV, nE)); }
  public SymBDD compose(SymBool oV, Object[] nE)    { return(comp(oV, nE)); }

  public SymBDD compose(SymBool[] oV, SymBool nE)   { return(comp(oV, nE)); }
//public SymBDD compose(SymBool[] oV, SymBool[] nE) { return(comp(oV, nE)); }
  public SymBDD compose(SymBool[] oV, SymUInt nE)   { return(comp(oV, nE)); }
  public SymBDD compose(SymBool[] oV, int[] nE)     { return(comp(oV, nE)); }
  public SymBDD compose(SymBool[] oV, Integer nE)   { return(comp(oV, nE)); }
  public SymBDD compose(SymBool[] oV, Object[] nE)  { return(comp(oV, nE)); }

  public SymBDD compose(SymUInt oV, SymBool nE)     { return(comp(oV, nE)); }
  public SymBDD compose(SymUInt oV, SymBool[] nE)   { return(comp(oV, nE)); }
  public SymBDD compose(SymUInt oV, SymUInt nE)     { return(comp(oV, nE)); }
  public SymBDD compose(SymUInt oV, int[] nE)       { return(comp(oV, nE)); }
  public SymBDD compose(SymUInt oV, Integer nE)     { return(comp(oV, nE)); }
  public SymBDD compose(SymUInt oV, Object[] nE)    { return(comp(oV, nE)); }

  public SymBDD compose(Object[] oV, SymBool nE)   { return(comp(oV, nE)); }
  public SymBDD compose(Object[] oV, SymUInt nE)   { return(comp(oV, nE)); }
  public SymBDD compose(Object[] oV, int[] nE)     { return(comp(oV, nE)); }
  public SymBDD compose(Object[] oV, Integer nE)   { return(comp(oV, nE)); }
  public SymBDD compose(Object[] oV, Object[] nE)  { return(comp(oV, nE)); }



  public SymBDD assign(SymBool oV, SymBool nE)     { return(assn(oV, nE)); }
  public SymBDD assign(SymBool oV, SymBool[] nE)   { return(assn(oV, nE)); }
  public SymBDD assign(SymBool oV, SymUInt nE)     { return(assn(oV, nE)); }
  public SymBDD assign(SymBool oV, int[] nE)       { return(assn(oV, nE)); }
  public SymBDD assign(SymBool oV, Integer nE)     { return(assn(oV, nE)); }
  public SymBDD assign(SymBool oV, Object[] nE)    { return(assn(oV, nE)); }

  public SymBDD assign(SymBool[] oV, SymBool nE)   { return(assn(oV, nE)); }
//public SymBDD assign(SymBool[] oV, SymBool[] nE) { return(assn(oV, nE)); }
  public SymBDD assign(SymBool[] oV, SymUInt nE)   { return(assn(oV, nE)); }
  public SymBDD assign(SymBool[] oV, int[] nE)     { return(assn(oV, nE)); }
  public SymBDD assign(SymBool[] oV, Integer nE)   { return(assn(oV, nE)); }
  public SymBDD assign(SymBool[] oV, Object[] nE)  { return(assn(oV, nE)); }

  public SymBDD assign(SymUInt oV, SymBool nE)     { return(assn(oV, nE)); }
  public SymBDD assign(SymUInt oV, SymBool[] nE)   { return(assn(oV, nE)); }
  public SymBDD assign(SymUInt oV, SymUInt nE)     { return(assn(oV, nE)); }
  public SymBDD assign(SymUInt oV, int[] nE)       { return(assn(oV, nE)); }
  public SymBDD assign(SymUInt oV, Integer nE)     { return(assn(oV, nE)); }
  public SymBDD assign(SymUInt oV, Object[] nE)    { return(assn(oV, nE)); }

  public SymBDD assign(Object[] oV, SymBool nE)   { return(assn(oV, nE)); }
  public SymBDD assign(Object[] oV, SymUInt nE)   { return(assn(oV, nE)); }
  public SymBDD assign(Object[] oV, int[] nE)     { return(assn(oV, nE)); }
  public SymBDD assign(Object[] oV, Integer nE)   { return(assn(oV, nE)); }
  public SymBDD assign(Object[] oV, Object[] nE)  { return(assn(oV, nE)); }


  public SymBDD assign(SymBool[] oldVars, SymBool[] newExprs) {
    // should be overridden by subclasses
    throw new UnsupportedOperationException();
  }
  public SymBDD assn(Object oV, Object nE) {
    List<SymBool> vList = new LinkedList<SymBool>();
    List<SymBool> eList = new LinkedList<SymBool>();
    comp(oV, nE, vList, eList);
    SymBool[] oldVars = vList.toArray(new SymBool[vList.size()]);
    SymBool[] newExprs = eList.toArray(new SymBool[eList.size()]);
    return(assign(oldVars, newExprs));
  }



  public boolean __setitem__(int i, boolean v) {
    __setitem__(i, mkSB(v));
    return(v);
  }
  public int __setitem__(int i, int v) {
    __setitem__(i, mkSB(v));
    return(v);
  }
  public long __setitem__(int i, long v) {
    __setitem__(i, mkSB(v));
    return(v);
  }
  public Object __setitem__(int i, SymBool v) {
    throw new UnsupportedOperationException();
  }
  public Object __setitem__(int i, Object v) {
    __setitem__(i, mkSB(v));
    return(v);
  }
  public boolean[] __setitem__(PySlice ps, boolean[] v) {
    __setitem__(ps, mkSBA(v));
    return(v);
  }
  public Object __setitem__(int i, SymBool[] v) {
    throw new UnsupportedOperationException();
  }
  public Object[] __setitem__(PySlice ps, Object[] v) {
    __setitem__(ps, mkSBA(v));
    return(v);
  }
  public int __setitem__(PySlice ps, int v) {
    __setitem__(ps, (long)(v));
    return(v);
  }
  public long __setitem__(PySlice ps, long v) {
    JavaSlice js = new JavaSlice(ps);
    SymBool[] x = new SymBool[js.indices(width()).size()];
    for(int i = 0; i < x.length; i++)
      x[i] = sf().bit(v, i);
    __setitem__(ps, x);
    return(v);
  }
  public SymUInt __setitem__(PySlice ps, SymUInt v) {
    __setitem__(ps, v.toArray());
    return(v);
  }
  public Object[] __setitem__(PySlice ps, SymBool[] v) {
    __setitem__(ps, mkSBA(v));
    return(v);
  }

  public boolean __nonzero__() {
    throw new UnsupportedOperationException("attempt to treat symbolic value as a non-symbolic boolean");
  }

  public SymBool mkSB(boolean b) { return(sf().bool(b)); }
  public SymBool mkSB(int i) { return(sf().bool((i & 1) == 1)); }
  public SymBool mkSB(long i) { return(sf().bool((i & 1) == 1)); }
  public SymBool mkSB(Object o) {
    if(o instanceof Boolean) return(mkSB(((Boolean)(o)).booleanValue()));
    else if(o instanceof Number) return(mkSB(((Number)(o)).longValue()));
    else if(o instanceof SymBool) return((SymBool)(o));
    else if(o instanceof SymUInt) return(((SymUInt)(o)).v(0));
    else if(o == null)
      throw new UnsupportedOperationException("mkSB(null)");
    else throw new UnsupportedOperationException("mkSB(" + o.getClass().getName() + " obj)");
  }
  public SymBool[] mkSBA(Object o) {
    if(o instanceof SymBool)
      return(new SymBool[] {(SymBool)(o)});
    else if(o instanceof SymUInt)
      return(((SymUInt)(o)).toArray());
    else if(o instanceof Boolean)
      return(new SymBool[] {sf().bool((Boolean)(o))});
    else if((o instanceof Number)) {
      long m = ((Number)(o)).longValue();
      SymBool[] a = new SymBool[width()];
      for(int i = 0; i < a.length; i++)
        a[i] = sf().bit(m, i);
      return(a);
    } else if(o.getClass().isArray()) {
      SymBool[] a = new SymBool[Array.getLength(o)];  
      for(int i = 0; i < a.length; i++)
        a[i] = mkSB(Array.get(o, i));
      return(a);
    } else throw new IllegalArgumentException();
  }
  public SymUInt mkSUI(Object o) {
    return(new SymUInt(mkSBA(o), sf()));
  }
  public SymBool[] getVars(Object obj, String who) {
    Set<SymBool> vSet = new TreeSet<SymBool>(new Comparator<SymBool>() {
      public int compare(SymBool a, SymBool b) {
        if(a.isVar() && b.isVar())
	  return(a.getBDD().level() - b.getBDD().level());
        else throw new IllegalArgumentException(
	  "variable expected (I should have detected this earlier)");
      }
    });
    getVars(obj, who, vSet);
    return(vSet.toArray(new SymBool[vSet.size()]));
  }
  public void getVars(Object obj, String who, Set<SymBool> vSet) {
    if(obj instanceof SymBool) {
      SymBool v = (SymBool)(obj);
      if(v.isVar()) vSet.add((SymBool)(obj));
      else
	throw new IllegalArgumentException(who
	  + ": argument not a variable -- got a " + obj.getClass());
    } else if(obj instanceof SymUInt) {
      getVars(((SymUInt)(obj)).toArray(), who, vSet);
    } else if(obj.getClass().isArray()) {
        for(int i = 0; i < Array.getLength(obj); i++)
	  getVars(Array.get(obj, i), who, vSet);
    } else {
      throw new IllegalArgumentException(who
        + ": argument not a variable -- got a " + obj.getClass());
    }
  }
}
