import net.sf.javabdd.*;
import java.util.*;

public class SymBool extends SymBDD {
  private int vnum;
  private BDD b;
  private int rank; // for BDD ordering
  public int rank() { return(rank); }

  public SymBool(String name, int rank, SymFactory sf) {
    super(name, 1, sf);
    vnum = sf().extVarNum(1); 
    this.b = sf().ithVar(vnum);
    this.rank = rank;
    sf().putVarSym(vnum, this);
  }
  public SymBool(String name, SymFactory sf) { this(name, 0, sf); }
  public SymBool(String name) { this(name, null); }
  public SymBool(BDD b, SymFactory sf) {
    super(1, sf);
    this.vnum = -1;
    this.rank = -1;
    this.b = b;
  }

  public static SymBool bool(boolean b, SymFactory sf) {
    return(sf.bool(b));
  }
  public static SymBool bool(boolean b) {
    return(bool(b, SymFactory.defaultSF()));
  }

  public BDD varBDD(SymBool[] vars, String who) {
    BDD b = sf().one().getBDD();
    for(int i = 0; i < vars.length; i++) {
      if(vars[i].isVar())
        b = b.and(vars[i].getBDD());
      else throw new IllegalArgumentException(
        who + "(vars): vars[" + i + "] is not a variable");
    }
    return(b);
  }

  // isVar(): true if the BDD is a single (non-negated) variable
  public boolean isVar() {
    BDD bdd = getBDD();
    if(bdd.isOne() || bdd.isZero()) return(false);
    return(bdd.high().isOne() && bdd.low().isZero());
  }

  public BDD getBDD() { // return my BDD
    sf().setVarOrderIfNeeded();
    return(b);
  }
  public int vnum() { // return my variable number in the BDD
    if(isVar()) return(vnum);
    else throw new UnsupportedOperationException(
    	"this SymBool is not a variable");
  }

  public SymUInt uint() { // uint: coinvert to a one-bit SymUInt 
    return(new SymUInt(new SymBool[] {this}, sf()));
  }

  // forwarding operations to methods of BDD, these are mostly in alphabetical
  // order.  The functions for compose() and equals are more than a single
  // one-liner.  They appear separately at the end of this list.
  public SymBool result(BDD v) {
    return(new SymBool(v, sf()));
  }
  public SymBool and(SymBool that) {
    return(result(this.getBDD().and(that.getBDD())));
  }
  public SymBool biimp(SymBool that) {
    return(result(this.getBDD().biimp(that.getBDD())));
  }
  public SymBool exist(SymBool[] vars) {
    return(result(getBDD().exist(varBDD(vars, "exist"))));
  }
  public SymBool forAll(SymBool[] vars) {
    return(result(getBDD().forAll(varBDD(vars, "forAll"))));
  }
  public SymBool high() {
    return(result(getBDD().high()));
  }
  public SymBool imp(SymBool that) {
    return(result(this.getBDD().imp(that.getBDD())));
  }
  public SymBDD ite(SymBool t, SymBool e) {
    return(result(this.getBDD().ite(t.getBDD(), e.getBDD())));
  }
  public double logSatCount() {
    return(getBDD().logSatCount()/Math.log(10.0));
  }
  public SymBool low() {
    return(result(getBDD().low()));
  }
  public SymBool neg() {
    return(this);  // for x in 0,1: -x & 1 = x, and we just compute the lsb
  }
  public int nodeCount() {
    return(getBDD().nodeCount());
  }
  public SymBool not() {
    return(result(this.getBDD().not()));
  }
  public SymBool or(SymBool that) {
    return(result(this.getBDD().or(that.getBDD())));
  }
  public SymBool relProd(SymBool that, SymBool[] vars) {
    return(result(this.getBDD().relprod(that.getBDD(), varBDD(vars, "relProd"))));
  }
  public double satCount() {
    return(getBDD().satCount());
  }
  public SymBool satOne() {
    BDD b = getBDD().satOne();
    if(b.isZero()) return(null);
    else return(new SymBool(b, sf()));
  }
  public SymBool topVar() {
    BDD b = getBDD();
    if(b.isZero() || b.isOne()) return(null);
    return(sf().varSym(sf().level2Var(b.level())));
  }
  public SymBool xor(SymBool that) {
    return(result(this.getBDD().xor(that.getBDD())));
  }
  public SymBool eq(SymBool that)  { return(biimp(that)); }
  public SymBool ge(SymBool that)  { return(that.imp(this)); }
  public SymBool gt(SymBool that)  { return(that.lt(this)); }
  public SymBool le(SymBool that)  { return(imp(that)); }
  public SymBool lt(SymBool that)  { return(not().and(that)); }
  public SymBool ne(SymBool that)  { return(xor(that)); }
  public SymBool shl(SymBool that) { return(gt(that)); }
  public SymBool shr(SymBool that) { return(gt(that)); }

  // compose provides the functionality of compose(), replace(), restrict(),
  //   and veccompose() from class BDD.  I check if the replacement expressions are
  //   all variables and use BDD.replace() instead of BDD.compose() guessing that
  //   BDD.replace() is more efficient.  I could do a similar check to see if
  //   BDD.restrict() could be used, but I'm guessing that there shouldn't be
  //   any significant performance difference.
  public SymBool compose(SymBool[] oldVars, SymBool[] newExprs) {
    if(oldVars.length != newExprs.length)
      throw new IllegalArgumentException(
          "mismatched lengths for oldVars and newExprs");
    BDDPairing pairs = sf().makePair();
    for(int i = 0; i < oldVars.length; i++) {
      if(!oldVars[i].isVar())
	throw new IllegalArgumentException("oldVar[" + i + "] is not a variable");
      pairs.set(oldVars[i].vnum(), newExprs[i].getBDD());
    }
    return(new SymBool(getBDD().veccompose(pairs), sf()));
  }

  public SymBool assign(SymBool[] oldVars, SymBool[] newExprs) {
    if(oldVars.length != newExprs.length)
      throw new IllegalArgumentException(
          "mismatched lengths for oldVars and newExprs");
    SymBool a = sf().one();
    for(int i = 0; i < oldVars.length; i++) {
      if(!oldVars[i].isVar())
	throw new IllegalArgumentException("oldVar[" + i + "] is not a variable");
      a = a.and(oldVars[i].eq(sf().bool(imp(newExprs[i]).equals(sf().one()))));
    }
    return(exist(oldVars).and(a));
  }


  // equals is almost a direct forward.  I provide overloaded versions for
  // comparisons with constants.
  public boolean equals(SymBool that) {
    return(this.getBDD().equals(that.getBDD()));
  }
  public boolean equals(boolean that) {
    if(that) return(this.getBDD().isOne());
    else return(this.getBDD().isZero());
  }
  public boolean equals(int that) {
    if(that == 0) return this.getBDD().isZero();
    else if(that == 1) return this.getBDD().isOne();
    else return(false);
  }


  // mapping operations from SymOp to the functions methods above
  public SymBDD op(SymBool that, SymOp op) {
    switch(op) {
      case ADD:  return(xor(that));
      case AND:  return(and(that));
      case  EQ:  return(biimp(that));
      case  GE:  return(ge(that));
      case  GT:  return(gt(that));
      case  LE:  return(imp(that));
      case  LT:  return(lt(that));
      case MUL:  return(and(that));
      case  NE:  return(ne(that));
      case  OR:  return(or(that));
      case SHL:  return(gt(that));
      case SHR:  return(gt(that));
      case SUB:  return(xor(that));
      case XOR:  return(xor(that));
      case RDIV: return(that.op(this, SymOp.DIV));
      case RMOD: return(that.op(this, SymOp.MOD));
      case RSHL: return(that.op(this, SymOp.SHL));
      case RSHR: return(that.op(this, SymOp.SHR));
      case RSUB: return(that.op(this, SymOp.SUB));
    }
    throw new UnsupportedOperationException(op.toString());
  }
  public SymBDD op(SymUInt that, SymOp op) {
    return(uint().op(that, op));
  }
  public SymBDD op(SymBDD that, SymOp op) {
    throw new UnsupportedOperationException(
      "SymBool: unfamiliar SymBDD: " + that.getClass());
  }
  public SymBDD op(int that, SymOp op) {
    return(op(sf().bit(that, 0), op));
  }


  // now stuff for printing SymBool values
  public static class CubeSet implements Iterable {
    private Set<BDD> cubes;
    private long maxCubes;
    private boolean truncated;

    public CubeSet(long maxCubes) {
      this.cubes = new HashSet<BDD>();
      this.maxCubes = maxCubes;
      this.truncated = false;
    }
    public CubeSet() { this(-1); }

    public boolean add(BDD b) {
      if((maxCubes >= 0) & (cubes.size() < maxCubes))
        cubes.add(b);
      else truncated = true;
      return(cubes.size() < maxCubes);
    }

    public Iterator<BDD> iterator() {
      return(cubes.iterator());
    }

    public int size() { return(cubes.size()); };

    public boolean contains(BDD b) { return(cubes.contains(b)); }

    public void remove(BDD b) { cubes.remove(b); }

    public boolean truncated() { return(truncated); }

    public void setTruncated() { truncated = true; }
  }

    
  public CubeSet cubes(BDD me, long maxCubes) {
    CubeSet c = new CubeSet(maxCubes);
    if(me.isOne()) c.add(me);
    else if(!me.isZero()) {
      int myVar = sf().level2Var(me.level());
      CubeSet loCubes = cubes(me.low(), maxCubes);
      if(!loCubes.truncated()) {
	CubeSet hiCubes = cubes(me.high(), maxCubes);
	Iterator<BDD> loIter = loCubes.iterator();
	cube1(loCubes, hiCubes, me.high(), c);
	cube1(hiCubes, loCubes, me.low(),  c);
	cube2(loCubes, sf().nithVar(myVar), c);
	cube2(hiCubes, sf().ithVar(myVar),  c);
	if(loCubes.truncated() | hiCubes.truncated())
	  c.setTruncated();
      } else {
	cube2(loCubes, sf().nithVar(myVar), c);
	c.setTruncated();
      }
    }
    return(c);
  }


  public void cube1(CubeSet s0, CubeSet s1, BDD b1, CubeSet c) {
    Iterator<BDD> it = s0.iterator();
    while(it.hasNext() && !c.truncated()) {
      BDD cube = it.next();
      if(s1.contains(cube)) {
        c.add(cube);
	it.remove();
	s1.remove(cube);
      } else if(cube.imp(b1).isOne()) {
        c.add(cube);
	it.remove();
      }
    }
  }

  public void cube2(CubeSet s0, BDD v, CubeSet c) {
    Iterator<BDD> it = s0.iterator();
    while(it.hasNext() && !c.truncated())
      c.add(v.and(it.next()));
  }

  public static class CubeComparator implements Comparator<BDD> {
    public int compare(BDD x, BDD y) {
      boolean xz = x.isZero(), xo = x.isOne();
      boolean yz = y.isZero(), yo = y.isOne();
      if(xz || xo || yz || yo) {
	int xv = (xo ? 1 : 0) + ((xz || xo) ? 0 : 2);
	int yv = (yo ? 1 : 0) + ((yz || yo) ? 0 : 2);
	return(xv-yv);
      }
      if(x.level() != y.level()) return(x.level() - y.level());
      boolean xTrue = x.low().isZero();
      boolean yTrue = x.low().isZero();
      if(xTrue == yTrue) {
	if(xTrue) return(compare(x.high(), y.high()));
	else return(compare(x.low(), y.low()));
      } else if(xTrue) return(1);
      else return(-1);
    }
  }

  public String toString(int maxCubes) {
    if(b.isOne()) return("True");
    else if(b.isZero()) return("False");
    else {
      TreeSet<BDD> sortCubes = new TreeSet<BDD>(new CubeComparator());
      StringBuffer buf = new StringBuffer(100);
      CubeSet c = cubes(b, maxCubes);
      sortCubes.addAll(c.cubes);

      Iterator<BDD> it = sortCubes.iterator();
      writeCube(buf, it.next()); // there must be at least one cube
      while(it.hasNext()) {
	buf.append(" | ");
	writeCube(buf, it.next());
      }
      if(c.truncated()) buf.append(" | ...");
      return(buf.toString());
    }
  }
  public String toString() { return(toString(100)); }

  public void writeCube(StringBuffer buf, BDD b) {
    BDD next;
    for(BDD x = b; !x.isOne(); x = next) {
      if(x != b) buf.append("&");
      if(x.low().isZero()) next = x.high();
      else if(x.high().isZero()) {
	buf.append("~");
	next = x.low();
      } else throw new IllegalArgumentException("SymBool.writeCube: not a cube");
      buf.append(sf().varName(sf().level2Var(x.level())));
    }
  }

  public boolean __nonzero__() {
    if(getBDD().isOne()) return(true);
    else if(getBDD().isZero()) return(false);
    else return(super.__nonzero__());
  }
}
