import net.sf.javabdd.*;
import java.util.*;
import org.python.core.PySlice;

public class SymUInt extends SymBDD implements Iterable<SymBool> {
  private SymBool[] bit, bit0;

  public SymUInt(String name, int nbits, SymFactory sf) {
    super(name, nbits, sf);
    this.bit  = new SymBool[nbits];
    this.bit0 = new SymBool[nbits];
    
    // Setting bit i at level -i puts the msb at the top of the BDD.
    // This allows way more BDD node sharing (for addition and related
    // operators) than the other way around.
    for(int i = 0; i < nbits; i++) {
      String name_i = name + "[" + i + "]";
      // If name[i] is already defined, we'll just use it.
      // This may cause some weird variable capture, but
      // I don't want to deal with cleaning up the symbol
      // tables in sf() if we've already made it this far.
      SymBool sym = (SymBool)(find(name_i));
      if(sym == null) sym = new SymBool(name_i, -i, sf());
      this.bit[i] = sym;
      this.bit0[i] = this.bit[i];
    }
  }
  public SymUInt(String name, int nbits) { this(name, nbits, null); }

  public SymUInt(SymBool[] bits, int nbits, SymFactory sf) {
    super(nbits, sf);
    init(bits);
  }
  public SymUInt(SymBool[] bits, SymFactory sf) { this(bits, bits.length, sf); }
  public SymUInt(SymBool[] bits, int nbits) { this(bits, nbits, null); }
  public SymUInt(SymBool[] bits) { this(bits, bits.length, null); }

  public SymUInt(SymBool bit, int nbits, SymFactory sf) {
    this(new SymBool[] {bit}, nbits, sf);
  }
  public SymUInt(SymBool bit, SymFactory sf) { this(bit, 1, sf); }
  public SymUInt(SymBool bit, int nbits) { this(bit, nbits, null); }
  public SymUInt(SymBool bit) { this(bit, 1, null); }

  public SymUInt(boolean[] bits, int nbits, SymFactory sf) {
    super(nbits, sf);
    init(bits);
  }
  public SymUInt(boolean[] bits, SymFactory sf) { this(bits, bits.length, sf); }
  public SymUInt(boolean[] bits, int nbits) { this(bits, nbits, null); }
  public SymUInt(boolean[] bits) { this(bits, bits.length, null); }

  public SymUInt(Boolean[] bits, int nbits, SymFactory sf) {
    super(nbits, sf);
    init(bits);
  }
  public SymUInt(Boolean[] bits, SymFactory sf) { this(bits, bits.length, sf); }
  public SymUInt(Boolean[] bits, int nbits) { this(bits, nbits, null); }
  public SymUInt(Boolean[] bits) { this(bits, bits.length, null); }

  public SymUInt(boolean bit, int nbits, SymFactory sf) {
    this(new boolean[] {bit}, nbits, sf);
  }
  public SymUInt(boolean bit, SymFactory sf) { this(bit, 1, sf); }
  public SymUInt(boolean bit, int nbits) { this(bit, nbits, null); }
  public SymUInt(boolean bit) { this(bit, 1, null); }

  public SymUInt(long m, int nbits, SymFactory sf) {
    super(nbits, sf);
    init(new Long(m));
  }
  public SymUInt(long m, int nbits) { this(m, nbits, null); }

  public SymUInt(Object[] bits, int nbits, SymFactory sf) {
    super(nbits, sf);
    init(bits);
  }
  public SymUInt(Object[] bits, SymFactory sf) { this(bits, bits.length, sf); }
  public SymUInt(Object[] bits, int nbits) { this(bits, nbits, null); }
  public SymUInt(Object[] bits) { this(bits, bits.length, null); }

  public SymUInt(SymUInt old, int nbits, SymFactory sf) {
    super(nbits, sf);
    init(old);
  }
  public SymUInt(SymUInt old, int nbits) {
    this(old, nbits, old.sf());
  }

  void init(Object obits) {
    SymBool[] bits = mkSBA(obits);
    this.bit  = new SymBool[width()];
    this.bit0 = new SymBool[width()];
    int w_min = Math.min(bits.length, width());
    for(int i = 0; i < w_min; i++)
      this.bit0[i] = bits[i];
    for(int i = w_min; i < width(); i++)
      this.bit0[i] = sf().zero();
    restore();
  }

  public SymBool v(int i) {
    return(bit[i]);
  }
  public SymBool v0(int i) {
    return(bit0[i]);
  }

  // true if every element of this SymUInt is a simple variable
  public boolean isVar() {
    if(name() != null) return(true);
    for(int i = 0; i < width(); i++)
      if(!v(i).isVar()) return(false);
    return(true);
  }

  public SymUInt resize(int w) {
    return(new SymUInt(this, w));
  }

  public SymUInt resizeVar(int w) {
    String nm = name();
    if(nm == null)
      throw new UnsupportedOperationException("'this' is not a variable");
    else if(w < 0)
      throw new IllegalArgumentException("resizeVar(w): w must be >= 0");
    SymBool[] bits = new SymBool[w];
    int w_min = Math.min(bits.length, width());
    for(int i = 0; i < w_min; i++)
      bits[i] = v(i);
    for(int i = w_min; i < w; i++) {
      String name_i = nm + "[" + i + "]";
      SymBool sym = (SymBool)(find(name_i));
      if(sym == null) sym = new SymBool(name_i, -i, sf());
      bits[i] = sym;
    }
    return(new SymUInt(bits, w, sf()));
  }

  public int nodeCount() {
    Vector<BDD> s = new Vector<BDD>(width());
    for(int i = 0; i < width(); i++)
      s.add(v(i).getBDD());
    return(sf().nodeCount(s));
  }

  public SymBDD[] addWithCarry(SymUInt that, SymBool carryIn) {
    int n0 = this.width();
    int n1 = that.width();
    int n_min = Math.min(n0, n1);
    int n_max = Math.max(n0, n1);
    SymBool carry = carryIn;
    SymBool sum[] = new SymBool[n_max];
    for(int i = 0; i < n_min; i++) {
      sum[i] = v(i).xor(that.v(i)).xor(carry);
      carry = v(i).and(that.v(i).or(carry)).or(that.v(i).and(carry));
    }
    SymUInt etc = (n0 < n1) ? this : that;
    for(int i = n_min; i < n_max; i++) {
      sum[i] = etc.v(i).xor(carry);
      carry  = etc.v(i).and(carry);
    }
    return(new SymBDD[] {new SymUInt(sum, sf()), carry});
  }

  public SymBDD add(SymUInt that) {
    return(addWithCarry(that, sf().zero())[0]);
  }

  public SymBDD and(SymUInt that) {
    if(this.width() != that.width())
      return(op(that, SymOp.AND));
    SymBool result[] = new SymBool[width()];
    for(int i = 0; i < width(); i++)
      result[i] = this.v(i).and(that.v(i));
    return(new SymUInt(result, sf()));
  }

  public SymBDD and() {
    SymBool x = sf().one();
    for(int i = 0; i < width(); i++)
      x = x.and(v(i));
    return(x);
  }

  public SymUInt compose(SymBool[] oldVars, SymBool[] newExprs) {
    SymBool result[] = new SymBool[width()];
    for(int i = 0; i < width(); i++)
      result[i] = v(i).compose(oldVars, newExprs);
    return(new SymUInt(result, sf()));
  }

  public SymBDD eq(SymUInt that) {
    return(ne(that).not());
  }

  public SymBool eq() {
    SymBool x = sf().one();
    for(int i = 1; i < width(); i++)
      x = x.and(v(i-1).eq(v(i)));
    return(x);
  }

  public boolean equals(SymUInt that) {
    if(this.width() != that.width())
      return(false);
    for(int i = 0; i < width(); i++)
      if(!this.v(i).equals(that.v(i)))
        return(false);
    return(true);
  }
  public boolean equals(boolean[] that) {
    return(equals(new SymUInt(that)));
  }
  public boolean equals(Boolean[] that) {
    return(equals(new SymUInt(that)));
  }
  public boolean equals(SymBool[] that) {
    return(equals(new SymUInt(that)));
  }
  public boolean equals(Object[] that) {
    return(equals(new SymUInt(that)));
  }

  public boolean equals() {
    boolean x = true;
    for(int i = 1; i < width(); i++)
      x = x && v(i-1).equals(v(i));
    return(x);
  }

  public SymUInt exist(SymBool[] vars) {
    SymBool[] x = new SymBool[width()];
    for(int i = 0; i < width(); i++)
      x[i] = v(i).exist(vars);
    return(new SymUInt(x, sf()));
  }

  public SymUInt forAll(SymBool[] vars) {
    SymBool[] x = new SymBool[width()];
    for(int i = 0; i < width(); i++)
      x[i] = v(i).forAll(vars);
    return(new SymUInt(x, sf()));
  }

  public SymBDD ge(SymUInt that) {
    SymBool x[] = that.lte(this);
    return(x[0].or(x[1]));
  }

  public SymBool ge() {
    SymBool x = sf().one();
    for(int i = 1; i < width(); i++)
      x = x.and(v(i-1).ge(v(i)));
    return(x);
  }

  public SymBDD gt(SymUInt that) {
    return(that.lte(this)[0]);
  }

  public SymBool gt() {
    SymBool x = sf().one();
    for(int i = 1; i < width(); i++)
      x = x.and(v(i-1).gt(v(i)));
    return(x);
  }

  public SymBDD le(SymUInt that) {
    SymBool x[] = this.lte(that);
    return(x[0].or(x[1]));
  }

  public SymBool le() {
    SymBool x = sf().one();
    for(int i = 1; i < width(); i++)
      x = x.and(v(i-1).le(v(i)));
    return(x);
  }

  public SymBDD lt(SymUInt that) {
    return(this.lte(that)[0]);
  }

  public SymBool lt() {
    SymBool x = sf().one();
    for(int i = 1; i < width(); i++)
      x = x.and(v(i-1).lt(v(i)));
    return(x);
  }

  public SymBDD ne(SymUInt that) {
    if(this.width() != that.width()) return(op(that, SymOp.NE));
    SymBool y = sf().zero();
    for(int i = 0; i < width(); i++)
      y = y.or(this.v(i).xor(that.v(i)));
    return(y);
  }

  public SymBool ne() {
    if(width() != 2) return(sf().zero());
    return(v(0).ne(v(1)));
  }

  public SymBDD neg() {
    return(not().op(1, SymOp.ADD));
  }

  public SymBDD not() {
    SymBool result[] = new SymBool[width()];
    for(int i = 0; i < width(); i++)
      result[i] = this.v(i).not();
    return(new SymUInt(result, sf()));
  }

  public SymBDD or(SymUInt that) {
    if(this.width() != that.width())
      return(op(that, SymOp.OR));
    SymBool result[] = new SymBool[width()];
    for(int i = 0; i < width(); i++)
      result[i] = this.v(i).or(that.v(i));
    return(new SymUInt(result, sf()));
  }

  public SymBDD or() {
    SymBool x = sf().zero();
    for(int i = 0; i < width(); i++)
      x = x.or(v(i));
    return(x);
  }

  public SymBDD sub(SymUInt that) {
    return(addWithCarry((SymUInt)(that.not()), sf().one())[0]);
  }

  public SymBDD xor(SymUInt that) {
    if(this.width() != that.width())
      return(op(that, SymOp.AND));
    SymBool result[] = new SymBool[width()];
    for(int i = 0; i < width(); i++)
      result[i] = this.v(i).xor(that.v(i));
    return(new SymUInt(result, sf()));
  }

  public SymBDD xor() {
    SymBool x = sf().zero();
    for(int i = 0; i < width(); i++)
      x = x.xor(v(i));
    return(x);
  }

  public SymBool[] lte(SymUInt that) {
    if(this.width() < that.width())
      return(resize(that.width()).lte(that));
    if(that.width() < this.width())
      return(lte(that.resize(this.width())));
    SymBool lt = sf().zero();  // less-than, 'so far'
    SymBool eq = sf().one();  // equal, 'so far'
    for(int i = width()-1; i >= 0; i--) {
      lt = lt.or(eq.and(this.v(i).not().and(that.v(i))));
      eq = eq.and(this.v(i).biimp(that.v(i)));
    }
    return(new SymBool[] {lt, eq});
  }

  // methods required by Iterable
  public Iterator<SymBool> iterator() {
    return(new Iterator<SymBool>() {
      int i = 0;
      public boolean hasNext() { return(i < width()); }
      public SymBool next() {
        if(i < width()) return(v(i++));
	else throw new NoSuchElementException();
      }
      public void remove() {
        throw new UnsupportedOperationException();
      }
    });
  }

  // functions for overloading python operators
  public int __len__() { return(width()); }

  public SymBool __getitem__(int i) {
    if((i < 0) || (i >= width()))
      throw new ArrayIndexOutOfBoundsException(
        "valid indices: 0.." + (width()-1) + ".  Got " + i);
    return(v(i));
  }
  public SymUInt __getitem__(PySlice ps) {
    Set<Integer> ii = (new JavaSlice(ps)).indices(width());
    SymBool[] x = new SymBool[ii.size()];
    int j = 0;
    for(int i : ii) x[j++] = v(i);
    return(new SymUInt(x, sf()));
  }

  public SymBool __setitem__(int i, SymBool v) {
    if((i < 0) || (i >= width()))
      throw new ArrayIndexOutOfBoundsException(
        "valid indices: 0.." + (width()-1) + ".  Got " + i);
    bit[i] = v;
    return(v);
  }
  public SymBool[] __setitem__(PySlice ps, SymBool[] v) {
    Set<Integer> ii = (new JavaSlice(ps)).indices(width());
    int j = 0;
    for(int i : ii) bit[i] = v[j++];
    return(v);
  }

  public SymBDD op(SymBool that, SymOp op) {
    return(op(that.uint(), op));
  }
  public SymBDD op(SymUInt that, SymOp op) {
    if(this.width() < that.width())
      return(resize(that.width()).op(that, op));
    if(that.width() < this.width())
      return(op(that.resize(this.width()), op));
    switch(op) {
      case ADD:  return(add(that));
      case AND:  return(and(that));
      case  EQ:  return(eq(that));
      case  GE:  return(ge(that));
      case  GT:  return(gt(that));
      case  LE:  return(le(that));
      case  LT:  return(lt(that));
      case  NE:  return(ne(that));
      case  OR:  return(or(that));
      case SUB:  return(sub(that));
      case XOR:  return(xor(that));
      case RSUB: return(that.op(this, SymOp.SUB));
    }
    throw new UnsupportedOperationException(op.toString());
  }
  public SymBDD op(int that, SymOp op) {
    SymBool x[] = new SymBool[width()];
    for(int i = 0; i < width(); i++)
      x[i] = sf().bit(that, i);
    return(op(new SymUInt(x, sf()), op));
  }

  public SymUInt copy() {
    return(new SymUInt(toArray(), sf()));
  }

  public SymUInt restore() {
    for(int i = 0; i < width(); i++)
      bit[i] = v0(i);
    return(this);
  }

  public SymBool[] toArray() {
    SymBool[] x = new SymBool[width()];
    for(int i = 0; i < x.length; i++)
      x[i] = v(i);
    return(x);
  }

  public String toString() {
    if(name() != null)
      return(name() + "[0:" + width() + "]");
    else {
      return("SymUInt@" + Integer.toHexString(hashCode()) + "[" + 0 + ":" + width() + "]");
    } 
  }

  public boolean __nonzero__() {
    return(or().__nonzero__());
  }
}
